/*!
 * This Source Code Form is subject to the terms of the Mozilla Public License,
 * v. 2.0. If a copy of the MPL was not distributed with this file, You can
 * obtain one at http://mozilla.org/MPL/2.0/.
 */

import { Schema as S } from "@effect/schema";
import type { Field } from "./Field";
import type { Method } from "./Method";
import type { MimeTypePredicate } from "./MimeType";
import { QueryEmpty, type Query } from "./Query";
import { RequestBodyFile, RequestBodyJson, RequestBodyMultipart, RequestBodyNone, RequestBodyText, RequestBodyUrlEncoded, type RequestBody } from "./RequestBody";
import { ResponseBodyFile, ResponseBodyJson, ResponseBodyText, ResponseEmpty, type ResponseBody } from "./ResponseBody";
import { Route, RouteLiteral, RouteParam } from "./Route";

export interface Api<Props extends Api.PropsAny> {
	readonly method: Method;
	readonly props: Props;
}

export namespace Api {
	export type Any = Api<PropsAny>;

	export type Props<T extends Api.Any> = T extends Api<infer Props> ? Props : never;

	export type PropsAny = {
		readonly route: Route.Any;
		readonly request: RequestBody.Any;
		readonly response: { readonly [_: number]: ResponseBody.Any };
		readonly query: Query.Any;
	};
}

export const Api = <const Props extends Api.PropsAny>(method: Method, props: Props): Api<Props> => Object.freeze<Api<Props>>({
	method,
	props,
});

// --- BUILDERS ----------------------------------------------------------------

export type RouteInputLiteral<Literal extends string> = Literal;

export type RouteInputParam<Name extends string, Schema extends S.Schema<any, string, unknown>> = readonly [Name, Schema];

export namespace RouteInputLiteral {
	export type Any = RouteInputLiteral<string>;
}

export namespace RouteInputParam {
	export type Any = RouteInputParam<string, S.Schema<any, string, unknown>>;
}

export namespace RouteInput {
	export type Any = readonly (
		| RouteInputLiteral.Any
		| RouteInputParam.Any
	)[];

	export type Route<R extends RouteInput.Any> = { readonly [K in keyof R]:
		R[K] extends RouteInputLiteral<infer Literal> ? RouteLiteral<Literal>
		: R[K] extends RouteInputParam<infer Name, infer Schema> ? RouteParam<Name, Schema>
		: never;
	};
}

export const make = <const R extends RouteInput.Any>(method: Method, ...routeInput: R): Api<{
	readonly route: RouteInput.Route<R>,
	readonly request: RequestBodyNone,
	readonly response: {},
	readonly query: {},
}> => Api(method, {
	route: Route(...routeInput.map((r) => typeof r === "string" ? RouteLiteral(r) : RouteParam(r[0], r[1])) as RouteInput.Route<R>),
	request: RequestBodyNone,
	response: ResponseEmpty,
	query: QueryEmpty,
});

export const requestBodyText = <const Schema extends S.Schema<any, string, unknown>>(schema: Schema) =>
	<const T extends Api.Any>({ method, props }: T): Api<{
		readonly route: T["props"]["route"],
		readonly request: RequestBodyText<Schema>,
		readonly response: T["props"]["response"],
		readonly query: T["props"]["query"],
	}> => Api(method, { ...props, request: RequestBodyText(schema) });

export const requestBodyJson = <const Schema extends S.Schema.Any>(schema: Schema) =>
	<const T extends Api.Any>({ method, props }: T): Api<{
		readonly route: T["props"]["route"],
		readonly request: RequestBodyJson<Schema>,
		readonly response: T["props"]["response"],
		readonly query: T["props"]["query"],
	}> => Api(method, { ...props, request: RequestBodyJson(schema) });

export const requestBodyUrlEncoded = <const Fields extends { readonly [_: string]: Field.AnyUrlEncoded }>(fields: Fields) =>
	<const T extends Api.Any>({ method, props }: T): Api<{
		readonly route: T["props"]["route"],
		readonly request: RequestBodyUrlEncoded<Fields>,
		readonly response: T["props"]["response"],
		readonly query: T["props"]["query"],
	}> => Api(method, { ...props, request: RequestBodyUrlEncoded(fields) });

export const requestBodyMultipart = <const Fields extends { readonly [_: string]: Field.Any }>(fields: Fields) =>
	<const T extends Api.Any>({ method, props }: T): Api<{
		readonly route: T["props"]["route"],
		readonly request: RequestBodyMultipart<Fields>,
		readonly response: T["props"]["response"],
		readonly query: T["props"]["query"],
	}> => Api(method, { ...props, request: RequestBodyMultipart(fields) });

export const requestBodyFile = <const MimeType extends string>(mimeType: MimeTypePredicate<MimeType>) =>
	<const T extends Api.Any>({ method, props }: T): Api<{
		readonly route: T["props"]["route"],
		readonly request: RequestBodyFile<MimeType>,
		readonly response: T["props"]["response"],
		readonly query: T["props"]["query"],
	}> => Api(method, { ...props, request: RequestBodyFile(mimeType) });

export const responseBodyText = <const Status extends number, const Schema extends S.Schema<any, string, unknown>>(status: Status, schema: Schema) =>
	<const T extends Api.Any>({ method, props: { response, ...props } }: T): Api<{
		readonly route: T["props"]["route"],
		readonly request: T["props"]["request"],
		readonly response: { readonly [K in keyof T["props"]["response"] | Status]: K extends Status ? ResponseBodyText<Schema> : T["props"]["response"][K] },
		readonly query: T["props"]["query"],
	}> => Api(method, { ...props, response: Object.freeze<any>({ ...response, [status]: ResponseBodyText(schema) }) });

export const responseBodyJson = <const Status extends number, const Schema extends S.Schema.Any>(status: Status, schema: Schema) =>
	<const T extends Api.Any>({ method, props: { response, ...props } }: T): Api<{
		readonly route: T["props"]["route"],
		readonly request: T["props"]["request"],
		readonly response: { readonly [K in keyof T["props"]["response"] | Status]: K extends Status ? ResponseBodyJson<Schema> : T["props"]["response"][K] },
		readonly query: T["props"]["query"],
	}> => Api(method, { ...props, response: Object.freeze<any>({ ...response, [status]: ResponseBodyJson(schema) }) });

export const responseBodyFile = <const Status extends number, const MimeType extends string>(status: Status, mimeType: MimeTypePredicate<MimeType>) =>
	<const T extends Api.Any>({ method, props: { response, ...props } }: T): Api<{
		readonly route: T["props"]["route"],
		readonly request: T["props"]["request"],
		readonly response: { readonly [K in keyof T["props"]["response"] | Status]: K extends Status ? ResponseBodyFile<MimeType> : T["props"]["response"][K] },
		readonly query: T["props"]["query"],
	}> => Api(method, { ...props, response: Object.freeze<any>({ ...response, [status]: ResponseBodyFile(mimeType) }) });

export const query = <const Query extends Query.Any>(query: Query) =>
	<const T extends Api.Any>({ method, props: { query: _, ...props } }: T): Api<{
		readonly route: T["props"]["route"],
		readonly request: T["props"]["request"],
		readonly response: T["props"]["response"],
		readonly query: Query,
	}> => Api(method, { ...props, query });
