/*!
 * This Source Code Form is subject to the terms of the Mozilla Public License,
 * v. 2.0. If a copy of the MPL was not distributed with this file, You can
 * obtain one at http://mozilla.org/MPL/2.0/.
 */

import { Schema as S } from "@effect/schema";
import type { MimeTypePredicate } from "./MimeType";

// --- INTERFACES --------------------------------------------------------------

export interface ResponseBodyNone {
	readonly _tag: "None";
}

export interface ResponseBodyText<Schema extends S.Schema<any, string, unknown>> {
	readonly _tag: "Text";
	readonly schema: Schema;
}

export interface ResponseBodyJson<Schema extends S.Schema.Any> {
	readonly _tag: "Json";
	readonly schema: Schema;
}

export interface ResponseBodyFile<MimeType extends string> {
	readonly _tag: "File";
	readonly mimeType: MimeTypePredicate<MimeType>;
}

export namespace ResponseBodyText {
	export type Any = ResponseBodyText<S.Schema<any, string, unknown>>;
}

export namespace ResponseBodyJson {
	export type Any = ResponseBodyJson<S.Schema.Any>;
}

export namespace ResponseBodyFile {
	export type Any = ResponseBodyFile<string>;
}

export namespace ResponseBody {
	export type Any =
		| ResponseBodyNone
		| ResponseBodyText.Any
		| ResponseBodyJson.Any
		| ResponseBodyFile.Any
		;
}

// --- CONSTRUCTORS ------------------------------------------------------------

export const ResponseBodyNone: ResponseBodyNone = Object.freeze<ResponseBodyNone>({
	_tag: "None",
});

export const ResponseBodyText = <const Schema extends S.Schema<any, string, unknown>>(schema: Schema): ResponseBodyText<Schema> => Object.freeze<ResponseBodyText<Schema>>({
	_tag: "Text",
	schema,
});

export const ResponseBodyJson = <const Schema extends S.Schema.Any>(schema: Schema): ResponseBodyJson<Schema> => Object.freeze<ResponseBodyJson<Schema>>({
	_tag: "Json",
	schema,
});

export const ResponseBodyFile = <const MimeType extends string>(mimeType: MimeTypePredicate<MimeType>): ResponseBodyFile<MimeType> => Object.freeze<ResponseBodyFile<MimeType>>({
	_tag: "File",
	mimeType,
});

export const ResponseEmpty: {} = Object.freeze<{}>({});
import type { Option as O } from "effect";

export type Multiplicity =
	| "single"
	| "optional"
	| "multiple"
	| "any"
	;

export namespace Multiplicity {
	export type With<M extends Multiplicity, T> = {
		single: T,
		optional: O.Option<T>,
		multiple: readonly [T, ...T[]],
		any: readonly T[],
	}[M];
}
